/*
This is part of ScreenVWM, which is based in part on the Litestep
shell source code.

Copyright (C) 2008 Jim Babcock
Copyright (C) 1997-2008 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#pragma once

/////////////////////////////////////////////////////////////////////////////

class NullElement
	:public LayoutElement
{
public:
	NullElement();
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	pair<int,int> getLength(ElementContext *context, bool vertical);
	void draw(HDC drawContext, LayoutCacheNode *layout);
};
NullElement *getNullElement();

/////////////////////////////////////////////////////////////////////////////

class RectLayoutElement
	:public LayoutElement
{
public:
	RectLayoutElement(string prefix);
	
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	pair<int,int> getLength(ElementContext *context, bool vertical);
	
	string x, y, width, height;
};

/////////////////////////////////////////////////////////////////////////////

class GroupElement
	:public LayoutElement
{
public:
	GroupElement(string prefix);
	~GroupElement();
	
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	pair<int,int> getLength(ElementContext *context, bool vertical);
	
	int baseLength;
	vector<LayoutElement*> children;
};

/////////////////////////////////////////////////////////////////////////////

class TextLabelElement
	:public LayoutElement
{
public:
	TextLabelElement(string prefix);
	~TextLabelElement();
	
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	pair<int,int> getLength(ElementContext *context, bool vertical);
	
	void draw(HDC drawContext, LayoutCacheNode *layout);
	bool changed(LayoutCacheNode *node);
	
protected:
	enum {
		labelAuto,
		labelDeskNum,
		labelTaskName,
		labelCustom,
	} text;
	string getText(ElementContext *context);
	
	Alignment alignHoriz;
	Alignment alignVert;
	string x, y;
	string minLength;
	string maxLength;
	string customText;
	PaintText *font;
	bool useStatsClass;
};

/////////////////////////////////////////////////////////////////////////////

class MinimapElement
	:public RectLayoutElement
{
public:
	MinimapElement(string prefix);
	void draw(HDC drawContext, LayoutCacheNode *layout);
	WindowData *getWindow(ElementContext *context, int px, int py);
	
	RECT screenToMinimapPos(ElementContext *context, Rect screenPos, POINT *iconPos=NULL);
	POINT minimapToScreenPos(ElementContext *context, int x, int y);
	
protected:
	friend class MinimapWindowPainter;
	
	int minimapTitleBarsThickness;
	int minimapIconSize;
	COLORREF minimapWindowColor;
	COLORREF minimapTitleBarColor;
	COLORREF minimapFocusedTitleBarColor;
	COLORREF minimapWindowBorderColor;
};

/////////////////////////////////////////////////////////////////////////////

class TextureElement
	:public RectLayoutElement
{
public:
	TextureElement(string prefix);
	~TextureElement();
	void draw(HDC drawContext, LayoutCacheNode *layout);
	
protected:
	PaintTexture *texture;
};

/////////////////////////////////////////////////////////////////////////////

class TaskIconElement
	:public RectLayoutElement
{
public:
	TaskIconElement(string prefix);
	void draw(HDC drawContext, LayoutCacheNode *layout);
	
protected:
	float offsetX, offsetY;
};

/////////////////////////////////////////////////////////////////////////////

class BranchElement
	:public LayoutElement
{
public:
	BranchElement(string prefix);
	~BranchElement();
	
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	pair<int,int> getLength(ElementContext *context, bool vertical);
	bool changed(LayoutCacheNode *node);
	
	struct BranchExpression {
		virtual bool evaluate(ElementContext *context)=0;
	};
	
protected:
	struct Branch {
		BranchExpression *condition;
		LayoutElement *element;
	};
	vector<Branch> branches;
	
	LayoutElement *getBranch(ElementContext *context);
	Branch parseBranch(const string &prefix);
	BranchExpression *parseExpression(const char *&pos);
	BranchExpression *parseUnaryExpression(const char *&pos);
	BranchExpression *parseLeafExpression(const string &expr);
};

/////////////////////////////////////////////////////////////////////////////

class SizeFallbackElement
	:public LayoutElement
{
public:
	SizeFallbackElement(string prefix);
	~SizeFallbackElement();
	// TODO
};

/////////////////////////////////////////////////////////////////////////////

class ChildModuleElement
	:public RectLayoutElement
{
public:
	typedef int (*ModuleInitExFunc)(HWND, HINSTANCE, LPCSTR);
	typedef int (*ModuleQuitFunc)(HINSTANCE);

	ChildModuleElement(string prefix);
	~ChildModuleElement();
	LayoutCacheNode *buildLayout(ElementContext *context, LayoutCacheNode *prev);
	void draw(HDC drawContext, LayoutCacheNode *layout);
	bool isModule() { return true; }
	
protected:
	FARPROC getModuleFunc(const char *name);
	void updateVars(bool doBangs, ElementContext *context);
	
	ModuleInitExFunc initFunc;
	ModuleQuitFunc quitFunc;
	
	string dllName;
	HINSTANCE dllInstance;
	HWND window;
	bool loaded;
	string modulePrefix;
	
	string moveCommand;
	string resizeCommand;
	string varX;
	string varY;
	string varWidth;
	string varHeight;
	
	int lastX, lastY;
	int lastWidth, lastHeight;
};

class ModuleLoadException
	{};
